/*
 * Mlex API Test.
 *
 * This file implements tests for the most important parts
 * of lexical analysis. Any other tests are easily to embed
 * here.
 *
 * Compile with:
 *      gcc -L../lib -I../include -I../lib  -Wall -g -o mlex_test \
 *             mlex_test.c mlex.o keyword.o symbol.o -lm -llib
 *
 */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "mlex.h"
#include "keyword.h"
#include "symbol.h"

#define ARRAY_SIZE(a)	(sizeof(a)/sizeof(a[0]))

#define RET_OK		0
#define RET_FAIL	1

#define INTEGER_MAX	2147483647
#define INTEGER_MIN	-2147483648
#define FLT_MAX         3.4028234663852886e+38
#define FLT_MIN         1.1754943508222875e-38

#define MSG(fmt, arg...) \
	do { fprintf(stderr, "mlex_test: " fmt "\n", ##arg); } while (0)

static int
test_number()
{
	int i,  ret;
	token_t token;
	char    lexstr[] = "128 -256 12.10 -12.10 +12.10 .10 +.10 "
			   "-.10 12.10e4 12.10e-4 12.e4 12.e-4";

	struct {
		token_t		token;
		union {
			int	int_val;
			float	float_val;
		};
	} table[] = {
		{ TOKEN_INTEGER, { .int_val   = 128 } },
		{ TOKEN_INTEGER, { .int_val   = -256 } },
		{ TOKEN_REAL,    { .float_val = 12.10f } },
		{ TOKEN_REAL,    { .float_val = -12.10f } },
		{ TOKEN_REAL,    { .float_val = 12.10f } },
		{ TOKEN_REAL,    { .float_val = 0.10f } },
		{ TOKEN_REAL,	 { .float_val = 0.10f } },
		{ TOKEN_REAL,    { .float_val = -0.10f } },
		{ TOKEN_REAL,    { .float_val = 12.10e4f } },
		{ TOKEN_REAL,    { .float_val = 12.10e-4f } },
		{ TOKEN_REAL,    { .float_val = 12.e4f } },
		{ TOKEN_REAL,    { .float_val = 12.e-4f } },
	};

	ret = RET_FAIL;

	mlex_file_setup(lexstr, lexstr+strlen(lexstr));

	for (i = 0; i < ARRAY_SIZE(table); i++) {

		token = mlex_get_next_token();

		if (token == TOKEN_EOF)
			break;

		if (table[i].token != token) {
			MSG("mlex_test: number test failed "
			    "at %d-th entry which was not "
			    "recognized as valid token type", i);
			goto out;
		}

		switch (token) {
		case TOKEN_INTEGER:
			if (table[i].int_val != mlex.integer) {
				MSG("mlex_test: number test failed "
				    "at %d-th entry which has invalid value "
				     "0x%08x should be 0x%08x",
				     i, mlex.integer, table[i].int_val);
				goto out;
			}
			break;
		case TOKEN_REAL:
			if (table[i].float_val != mlex.real) {
				MSG("mlex_test: number test failed "
				    "at %d-th entry which has invalid "
				    "floating point value", i);
				goto out;
			}
			break;
		default:
			/* Shouldn't reach. */
			break;
		}
	}

	ret = RET_OK;
out:
	mlex_file_finalize();
	return ret;
}

static int
test_integer_overflow()
{
	char  lexstr[] = "2147483648 -2147483649";
	int   limits[] = { INTEGER_MAX, INTEGER_MIN };
	char *limstr[] = { "INTEGER_MAX", "INTEGER_MIN" };
	int i, ret;
	token_t token;

	ret = RET_FAIL;

	mlex_file_setup(lexstr, lexstr + strlen(lexstr));

	for (i = 0; i < 2; i++) {

		token = mlex_get_next_token();

		if (token != TOKEN_INTEGER) {
			MSG("mlex_test: integer overflow test "
			    "failed, couldn't recognize %s token",
			    limstr[i]);
			goto out;
		}

		if (mlex.integer != limits[i]) {
			MSG("mlex_test: integer overflow test "
			    "failed, overflow of %s not detected",
			    limstr[i]);
			goto out;
		}
	}

	ret = RET_OK;
out:
	mlex_file_finalize();
	return ret;
}

static int
test_integer_max()
{
	char  lexstr[] = "2147483647 -2147483648";
	char *intstr[] = { "INTEGER_MAX", "INTEGER_MIN" };
	int   maxint[] = { INTEGER_MAX, INTEGER_MIN };
	token_t token;
	int i, ret;

	ret = RET_FAIL;

	mlex_file_setup(lexstr, lexstr + strlen(lexstr));

	for (i = 0; i < 2; i++) {

		token = mlex_get_next_token();

		if (token != TOKEN_INTEGER) {
			MSG("mlex_test: integer max test failed, "
			    "couldn't recognize %s token", intstr[i]);
			goto out;
		}

		if (mlex.integer != maxint[i]) {
			MSG("mlex_test: integer max test failed, "
			    "on token %s, got %d expected %d",
			    intstr[i], mlex.integer, maxint[i]);
			goto out;
		}	
	}

	ret = RET_OK;
out:
	mlex_file_finalize();
	return ret;
}

/*
 * TODO: test overflow/underflow when double value in mlex subjects it.
 */
static int
test_float_overflow_underflow()
{
	char  lexstr[] = "3.40282357e+38 -3.40282357e+38 "
			 "1.17549428e-38 -1.17549428e-38";
	float maxflt[] = { FLT_MAX, -FLT_MAX, FLT_MIN, -FLT_MIN };
	char *fltstr[] = { "+FLT_MAX", "-FLT_MAX", "FLT_MIN", "-FLT_MIN" };
	int  fpclass[] = { FP_INFINITE, FP_INFINITE, FP_SUBNORMAL, FP_SUBNORMAL };
	token_t token;
	int i, ret, c;

	ret = RET_FAIL;

	mlex_file_setup(lexstr, lexstr + strlen(lexstr));

	for (i = 0; i < ARRAY_SIZE(fpclass); i++) {

		token = mlex_get_next_token();

		if (token != TOKEN_REAL) {
			MSG("mlex_test: float overflow/underflow "
			    "test failed, couldn't recognize %s token",
			    fltstr[i]);
			goto out;
		}

		c = fpclassify(mlex.real);

		if (c == FP_INFINITE) {
			if (maxflt[i] < 0.0 && mlex.real != -INFINITY) {
				MSG("mlex_test: float "
				    "overflow/underflow test failed, "
				    "%s is not -infinity", fltstr[i]);
				goto out;
			} else if (maxflt[i] > 0.0 && mlex.real != INFINITY) {
				MSG("mlex_test: float overflow/underflow "
				    "test failed, %s is not +infinity",
				    fltstr[i]);
				goto out;
			}
		}

		if (c != fpclass[i]) {
			MSG("mlex_test: float overflow/underflow "
			    "test failed, couldn't detect over/underflow "
			    "with value %s %s",
			    maxflt[i] == FLT_MAX ? ">" : "<", fltstr[i]);
			goto out;
		}
	}

	ret = RET_OK;
out:
	mlex_file_finalize();
	return ret;
}

#define SPCHR(arg)	"'\\" #arg "'"

static int
test_special_character()
{
	char lexstr[] = SPCHR(0) SPCHR(a) SPCHR(b) SPCHR(f) SPCHR(n) SPCHR(r)
			SPCHR(t) SPCHR(v) "'\\\\'" "'\"'" SPCHR(?) SPCHR(01)
			SPCHR(001) SPCHR(0001) SPCHR(0777) SPCHR(0x1)
			SPCHR(0x01) SPCHR(0xff);
	char     ch[] = { '\0', '\a', '\b', '\f', '\n', '\r', '\t', '\v',
	                  '\\', '\"', '\?', 0x1, 0x1, 0x1, 0xff, 0x1,
		          0x1, 0xff };
	int i, ret;
	token_t token;

	ret = RET_FAIL;

	mlex_file_setup(lexstr, lexstr + strlen(lexstr));

	for (i = 0; i < ARRAY_SIZE(ch); i++) {

		token = mlex_get_next_token();

		if (token != TOKEN_CHAR) {
			MSG("mlex_test: special character "
			    "test failed, couldn't parse %d-th token "
			    "as TOKEN_CHAR", i);
			goto out;
		}

		if (mlex.ch != ch[i]) {
			MSG("mlex_test: special character "
			    "test failed, character values at %d-th "
			    "index do not match", i);
			goto out;
		}
	}

	ret = RET_OK;
out:
	mlex_file_finalize();
	return ret;
}

static int
test_special_character_overflow()
{
	char lexstr[] = "'\\004777''\\0x1ff'";
	int  intval[] = { 255, 255 };
	int i, ret;
	token_t token;

	ret = RET_FAIL;

	mlex_file_setup(lexstr, lexstr + strlen(lexstr));

	for (i = 0; i < ARRAY_SIZE(intval); i++) {

		token = mlex_get_next_token();

		if (token != TOKEN_CHAR) {
			MSG("mlex_test: special character "
			    "overflow test failed, couldn't parse %d-th "
			    "token as TOKEN_CHAR", i);
			goto out;
		}

		if (mlex.ch != (char) intval[i]) {
			MSG("mlex_test: special character "
			    "overflow test failed, %d-th overflowed value "
			    "is invalid, got %0x02x expected %0x02x",
			     i, mlex.ch, intval[i]);
			goto out;
		}
	}

	ret = RET_OK;
out:
	mlex_file_finalize();
	return ret;
}

static int
test_all_tokens()
{
	char all_tokens[] = "128 0.1 'a' \"hello!\\n\" _my_cool_id99 "
			"*/+-%$;,{}[]()=<><=>=!===||&&!."
			"int float char string void if else while for "
			"break continue return module";

	int tlist[] = { TOKEN_INTEGER, TOKEN_REAL, TOKEN_CHAR, TOKEN_STRING,
			TOKEN_ID, TOKEN_ASTERISK, TOKEN_SLASH, TOKEN_PLUS,
			TOKEN_MINUS, TOKEN_PERCENT, TOKEN_DOLLAR, TOKEN_SEMICOLON,
			TOKEN_COMMA, TOKEN_LBRACE, TOKEN_RBRACE, TOKEN_LBRACKET,
			TOKEN_RBRACKET, TOKEN_LPARENTH, TOKEN_RPARENTH, TOKEN_EQUALITY,
			TOKEN_LT, TOKEN_GT, TOKEN_LE, TOKEN_GE,
			TOKEN_NE, TOKEN_EQ, TOKEN_OR, TOKEN_AND,
			TOKEN_NOT, TOKEN_DOT, TOKEN_DECL_INT, TOKEN_DECL_REAL,
			TOKEN_DECL_CHAR, TOKEN_DECL_STRING, TOKEN_DECL_VOID, TOKEN_IF,
			TOKEN_ELSE, TOKEN_WHILE, TOKEN_FOR, TOKEN_BREAK,
			TOKEN_CONTINUE, TOKEN_RETURN, TOKEN_MODULE };
	int i, ret;
	token_t token;

	ret = RET_FAIL;

	symbol_table_create_global_table();

	keyword_table_create();

	mlex_file_setup(all_tokens, all_tokens + strlen(all_tokens));

	for (i = 0; i < ARRAY_SIZE(tlist); i++) {

		token = mlex_get_next_token();

		if ((token != tlist[i]) || (mlex.token != tlist[i])) {
			MSG("mlex_test: all tokens test failed at "
			    "%d-th item", i);
			goto out;
		}
	}

	ret = RET_OK;
out:
	keyword_table_destroy();
	mlex_file_finalize();
	return ret;
}

static int
test_string_constant()
{
	char lexstr[] = "\"the quick brown fox jumps over the lazy dog\""
			"\"\\a\\b\\f\\n\\r\\t\\v\""
			"\"\\0\"";
	char *str[] = {
		"the quick brown fox jumps over the lazy dog",
  		"\\a\\b\\f\\n\\r\\t\\v",
		"\\0"
	};
	int i, ret;
	token_t token;

	ret = RET_FAIL;
	
	mlex_file_setup(lexstr, lexstr + strlen(lexstr));

	for (i = 0; i < ARRAY_SIZE(str); i++) {

		token = mlex_get_next_token();

		if (token != TOKEN_STRING) {
			MSG("mlex_test: constant string test "
			    "failed, couldn't to recognize %d-th item as "
			    "TOKEN_STRING", i);
			goto out;
		}

		if (strcmp(mlex.str, str[i]) != 0) {
			MSG("mlex_test: constant string test "
			    "failed, %d-th item strings do not match", i);
			goto out;
		}
	}

	ret = RET_OK;
out:
	mlex_file_finalize();
	return ret;
}

static int
test_missing_paired_quote()
{
	char lexstr[] = { '\'', '\\', '0', '0', '0', '\n',
			  '\'', '\\', '0', '0', '0', '1', '\'',
			  '\0' };
	int ret, line;
	token_t token;

	ret = RET_FAIL;

	mlex_file_setup(lexstr, lexstr + strlen(lexstr));

	line = mlex_line_number();

	token = mlex_get_next_token();

	if (token != TOKEN_UNKNOWN) {
		MSG("missing paired quote test failed because "
		    "ivalid sequence was not detected");
		goto out;
	}

	while (mlex_get_next_token() != TOKEN_EOF)
		;

	if (mlex_line_number() != 1) {
		MSG("missing paired quote test results in "
		    "invalid line number after recovery");
		goto out;
	}

	ret = RET_OK;
out:
	mlex_file_finalize();
	return ret;
}

static int
test_charconst_multiple()
{
	char   lexstr[] = "'db' 'ea' 'ab b' 'dy '";
	char const ch[] = { 'b', 'a', 'b', ' ' };
	token_t token;
	int ret, i;

	ret = RET_FAIL;

	mlex_file_setup(lexstr, lexstr + strlen(lexstr));

	for (i = 0; i < ARRAY_SIZE(ch); i++) {

		token = mlex_get_next_token();

		if (token != TOKEN_CHAR) {
			MSG("multiple character test failed, couldn't "
			    "detect %d-th token as TOKEN_CHAR", i);
			goto out;
		}

		if (mlex.ch != ch[i]) {
			MSG("multiple character test failed, because "
			    "%d-th token doesn't match, "
			    "got 0x%02x expected 0x%02x",
			    i, mlex.ch, ch[i]);
			goto out;
		}
	}

	ret = RET_OK;
out:
	mlex_file_finalize();
	return ret;
}

typedef int (*test_callback_t)(void);

int
mlex_test()
{
	int i, ret;

	test_callback_t testcbs[] = {
		test_number,
		test_integer_overflow,
		test_integer_max,
		test_float_overflow_underflow,
		test_special_character,
		test_special_character_overflow,
		test_all_tokens,
		test_string_constant,
		test_missing_paired_quote,
		test_charconst_multiple
	};

	for (i = 0; i < ARRAY_SIZE(testcbs); i++) {

		ret = (*testcbs[i])();

		if (ret != 0)
			break;
	}

	return ret;
}

int
main(int argc, char *argv[])
{
	int res;
	char *msg[] = { "ok", "failed" };

	res = mlex_test();

	printf("mlex_test: %s\n", msg[res]);

	return res;
}